static char *rcsid= "$Id: glwin.c,v 2.84 2003/03/28 16:47:48 fuhrmann Exp $";
/*

  TODO:
  pthd: properly handle dump requests from main thread
  pthd: problem on sgi with fonts after win delete
  pthd: sync without blocking  -- try XSendEvent again ???
  pthd: replace #ifdefs by if (glw_threaded) and wrap all pthread calls.
  pthd: figure out from which thread we are called  -- like in setfontsize stuff
  pthd: signal handling
 */


#ifdef NO_FLTK
#undef FLTK
#endif
#ifndef FLTK
#undef PTHREADS

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <GL/glx.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <assert.h>
#include <time.h>

#ifdef PTHREADS
#include <pthread.h>
#endif

#include "glwin.h"
#include "gleps.h" /* TODO: avoid this depedency somehow */

#define GLW_BUFSIZE 256
#define GLW_MAX_FONT_SIZE 50
#define GLW_FONT_NAME  "-*-*-*-r-normal-*-%d-*-*-*-m-*-iso8859-1"
#define GLW_DEFAULT_FONT_SIZE 0.025
static int debug=0;
static char glw_string_buffer[GLW_BUFSIZE];

#define glwTrace(w,x) {if (debug) fprintf(stdout,"glwin.c(%d): %p %s\n",__LINE__,(void*)w,x);}
void glwDebug(void)
{
  debug=!debug;
  if (debug) 
    fprintf(stderr,"glwin: debugging enabled.\n");
  else
    fprintf(stderr,"glwin: debugging disabled.\n");
}

#define glwError(msg) glw_error(__LINE__,msg) 
static void glw_error(int line, char * msg)
{
 fprintf(stderr,"\nglwin.c: error on line %d: %s\n",line,msg);
}

static int ErrorHandler(Display *display, XErrorEvent *event)
{
  char buf[80];
  glwError("Received X error!");
  fprintf(stderr," Error code   : %d\n", event->error_code);
  fprintf(stderr," Request code : %d\n", event->request_code);
  fprintf(stderr," Minor code   : %d\n", event->minor_code);
  XGetErrorText(display, event->error_code, buf, 80);
  fprintf(stderr," Error text   : %s\n", buf);
  return 0;
}

struct glWindowStruct 
{
  int x, y, w, h;
  GLenum type;
  GLenum dm_policy;
  Window w_main, w_overlay, w_toplevel;
  Pixmap pixmap;
  

  XVisualInfo *vInfoMain, *vInfoOverlay;
  int vInfoMainClass;
  int vInfoOverlayClass;
  Colormap cMapMain, cMapOverlay;
  GLXPixmap glx_pixmap;
  GLXContext cMain, cOverlay;
  
  void (*redraw_func)(glWindow, void *);
  void (*expose_func)(glWindow, void *,int, int);
  void (*reshape_func)(glWindow, void *,int, int);
  GLenum (*key_down_func)(glWindow, void *,int, GLenum);
  GLenum (*mouse_down_func)(glWindow, void *,int, int, GLenum);
  GLenum (*mouse_up_func)(glWindow, void *,int, int, GLenum);
  GLenum (*mouse_move_func)(glWindow, void *,int, int, GLenum);
  int lastEventType;
  GLenum drawAllowFlag;
  GLenum overlayFlag;
  
  Atom deleteWindowAtom;
  char *title,*xtitle,*state;
  void * info;

  GLuint font_base[GLW_MAX_FONT_SIZE];
  GLuint font_range[GLW_MAX_FONT_SIZE];
  double font_size;
  int font_to_use;
  int keep_aspect;

  Cursor wait_cursor;
  Cursor user_cursor;

  int offscreen;
  int swap;
  int firstswap;

  int application_redraw_request;
  int event_redraw_request;
  int destroy_request;
  int create_request;
  glWindow next;
#include "gldump.h"
};
#include "gldump.c"

#ifdef PTHREADS
static   pthread_mutex_t glw_mutex, glw_cond_mutex,glw_precond_mutex;
static   pthread_cond_t glw_cond, glw_precond;
static   pthread_attr_t glw_attr;
static   pthread_t glw_thread;
#endif

static int  glw_threads=-1;

static volatile glWindow glw_anchor=NULL;
static glWindow glw_current=NULL;
static Display *glw_display=NULL;
static int glw_screen=0; 
static Window glw_root=0;
static int glw_control_mode=GLW_APPLICATION_DRIVEN;
static int glw_process_events=0;
static XFontStruct *glw_font_info[GLW_MAX_FONT_SIZE];

int glwEnableThreads(void)
{
#ifdef PTHREADS
  if (glw_threads==-1)
    {
      glw_threads=1;
      return 1;
    }
  else
    {
      glwError("glwEnableThreads has to be called befor any other glwin function\n");
      return 0;
    }
#else
  glwError("PTHREADS not defined during compilation");
  return 0;
#endif
}
int glwDisableThreads(void)
{
#ifdef PTHREADS
  if (glw_threads==-1)
    {
      glw_threads=0;
      return 1;
    }
  else
    {
      glwError("glwDisableThreads has to be called befor any other glwin function\n");
      return 0;
    }
#else
  return 1;
#endif
}


/******************************************************************************/
static void glw_next_event(void);

static void glw_insert(glWindow w)
{
  glWindow wx;
  glwLock();
  if (glw_anchor==NULL)
    glw_anchor=w;
  else
    {
      for (wx=glw_anchor;wx->next!=NULL;wx=wx->next);
      wx->next=w;
    }
  w->next=NULL;
  glwUnlock();
}

static void glw_remove(glWindow w)
{
  glWindow wx;
  glwLock();
  if (w==glw_anchor)
    glw_anchor=w->next;
  else
    {
      for (wx=glw_anchor;wx->next!=w;wx=wx->next);
      assert(wx->next==w);
      wx->next=w->next;
    }
  glwUnlock();
}

static glWindow glw_find(Window w)
{
  glWindow wx;
  for (wx=glw_anchor;wx&&wx->w_main!=w;wx=wx->next);
  return wx;
}
/******************************************************************************/

float colorMaps[] = {
    0.000000, 1.000000, 0.000000, 1.000000, 0.000000, 1.000000, 
    0.000000, 1.000000, 0.333333, 0.776471, 0.443137, 0.556863, 
    0.443137, 0.556863, 0.219608, 0.666667, 0.666667, 0.333333, 
    0.666667, 0.333333, 0.666667, 0.333333, 0.666667, 0.333333, 
    0.666667, 0.333333, 0.666667, 0.333333, 0.666667, 0.333333, 
    0.666667, 0.333333, 0.039216, 0.078431, 0.117647, 0.156863, 
    0.200000, 0.239216, 0.278431, 0.317647, 0.356863, 0.400000, 
    0.439216, 0.478431, 0.517647, 0.556863, 0.600000, 0.639216, 
    0.678431, 0.717647, 0.756863, 0.800000, 0.839216, 0.878431, 
    0.917647, 0.956863, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 0.749020, 0.749020, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 1.000000, 1.000000, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.749020, 0.749020, 
    0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 
    1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 
    1.000000, 1.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 0.749020, 0.749020, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 1.000000, 1.000000, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 0.000000, 0.000000, 
    1.000000, 1.000000, 0.000000, 0.000000, 1.000000, 1.000000, 
    0.333333, 0.443137, 0.776471, 0.556863, 0.443137, 0.219608, 
    0.556863, 0.666667, 0.666667, 0.333333, 0.666667, 0.333333, 
    0.666667, 0.333333, 0.666667, 0.333333, 0.666667, 0.333333, 
    0.666667, 0.333333, 0.666667, 0.333333, 0.666667, 0.333333, 
    0.039216, 0.078431, 0.117647, 0.156863, 0.200000, 0.239216, 
    0.278431, 0.317647, 0.356863, 0.400000, 0.439216, 0.478431, 
    0.517647, 0.556863, 0.600000, 0.639216, 0.678431, 0.717647, 
    0.756863, 0.800000, 0.839216, 0.878431, 0.917647, 0.956863, 
    0.000000, 0.141176, 0.282353, 0.427451, 0.568627, 0.713726, 
    0.854902, 1.000000, 0.000000, 0.141176, 0.282353, 0.427451, 
    0.568627, 0.713726, 0.854902, 1.000000, 0.000000, 0.141176, 
    0.282353, 0.427451, 0.568627, 0.713726, 0.854902, 1.000000, 
    0.000000, 0.141176, 0.282353, 0.427451, 0.568627, 0.713726, 
    0.854902, 1.000000, 0.000000, 0.141176, 0.282353, 0.427451, 
    0.568627, 0.713726, 0.854902, 1.000000, 0.000000, 0.141176, 
    0.282353, 0.427451, 0.568627, 0.713726, 0.854902, 1.000000, 
    0.000000, 0.141176, 0.282353, 0.427451, 0.568627, 0.713726, 
    0.854902, 1.000000, 0.000000, 0.141176, 0.282353, 0.427451, 
    0.568627, 0.713726, 0.854902, 1.000000, 0.000000, 0.141176, 
    0.282353, 0.427451, 0.568627, 0.713726, 0.854902, 1.000000, 
    0.000000, 0.141176, 0.282353, 0.427451, 0.568627, 0.713726, 
    0.854902, 1.000000, 0.000000, 0.141176, 0.282353, 0.427451, 
    0.568627, 0.713726, 0.854902, 1.000000, 0.000000, 0.141176, 
    0.282353, 0.427451, 0.568627, 0.713726, 0.854902, 1.000000, 
    0.000000, 0.141176, 0.282353, 0.427451, 0.568627, 0.713726, 
    0.854902, 1.000000, 0.000000, 0.141176, 0.282353, 0.427451, 
    0.568627, 0.713726, 0.854902, 1.000000, 0.000000, 0.141176, 
    0.282353, 0.427451, 0.568627, 0.713726, 0.854902, 1.000000, 
    0.000000, 0.141176, 0.282353, 0.427451, 0.568627, 0.713726, 
    0.854902, 1.000000, 0.000000, 0.141176, 0.282353, 0.427451, 
    0.568627, 0.713726, 0.854902, 1.000000, 0.000000, 0.141176, 
    0.282353, 0.427451, 0.568627, 0.713726, 0.854902, 1.000000, 
    0.000000, 0.141176, 0.282353, 0.427451, 0.568627, 0.713726, 
    0.854902, 1.000000, 0.000000, 0.141176, 0.282353, 0.427451, 
    0.568627, 0.713726, 0.854902, 1.000000, 0.000000, 0.141176, 
    0.282353, 0.427451, 0.568627, 0.713726, 0.854902, 1.000000, 
    0.000000, 0.141176, 0.282353, 0.427451, 0.568627, 0.713726, 
    0.854902, 1.000000, 0.000000, 0.141176, 0.282353, 0.427451, 
    0.568627, 0.713726, 0.854902, 1.000000, 0.000000, 0.141176, 
    0.282353, 0.427451, 0.568627, 0.713726, 0.854902, 1.000000, 
    0.000000, 0.141176, 0.282353, 0.427451, 0.568627, 0.713726, 
    0.854902, 1.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 0.333333, 0.443137, 
    0.443137, 0.219608, 0.776471, 0.556863, 0.556863, 0.666667, 
    0.666667, 0.333333, 0.666667, 0.333333, 0.666667, 0.333333, 
    0.666667, 0.333333, 0.666667, 0.333333, 0.666667, 0.333333, 
    0.666667, 0.333333, 0.666667, 0.333333, 0.039216, 0.078431, 
    0.117647, 0.156863, 0.200000, 0.239216, 0.278431, 0.317647, 
    0.356863, 0.400000, 0.439216, 0.478431, 0.517647, 0.556863, 
    0.600000, 0.639216, 0.678431, 0.717647, 0.756863, 0.800000, 
    0.839216, 0.878431, 0.917647, 0.956863, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 
    0.000000, 0.000000, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 0.247059, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 0.498039, 
    0.498039, 0.498039, 0.498039, 0.498039, 0.749020, 0.749020, 
    0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 0.749020, 
    0.749020, 0.749020, 1.000000, 1.000000, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 
    1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 1.000000, 
};
float glwRGBMap[8][3] = {
    {
	0, 0, 0
    },
    {
	1, 0, 0
    },
    {
	0, 1, 0
    },
    {
	1, 1, 0
    },
    {
	0, 0, 1
    },
    {
	1, 0, 1
    },
    {
	0, 1, 1
    },
    {
	1, 1, 1
    }
};

/******************************************************************************/


#if defined(DEC) && defined(vms)

static int ffs (unsigned int mask)
{
    int			num;
    unsigned int	bit;

    if (mask == 0) return 0;
    
    for (num = 1, bit = 1; (mask & bit) != 0; num++, bit <<= 1) {}

    return num;

}

#endif /* DEC */
/******************************************************************************/

int glwGetColorMapSize(glWindow w)
{

  return w->vInfoMain->colormap_size;
    
}
/******************************************************************************/

static int glwLoadFont(glWindow w,int isize)
{
  Font id;
  int first, last;
  char fontname[128];
  glwAttach(w);
  if (debug) 
    {
      fprintf(stderr,"Load font size %d\n",isize);

    }
  if ((isize<0)&& (isize>=GLW_MAX_FONT_SIZE))
    {
      glwError("font size out of range\n");
      return 0;
    }
  if (glw_font_info[isize] == NULL) 
    {
      sprintf(fontname,GLW_FONT_NAME,isize);
      glw_font_info[isize] = XLoadQueryFont(glw_display, fontname);
      if (glw_font_info[isize] == NULL) 
	{
	  glwError("font not found\n");
	  sprintf(fontname,GLW_FONT_NAME,12);
	  glw_font_info[isize] = XLoadQueryFont(glw_display, fontname);
	}
    }
  id = glw_font_info[isize]->fid;
  first = glw_font_info[isize]->min_char_or_byte2;
  last = glw_font_info[isize]->max_char_or_byte2;
  
  w->font_base[isize] = glGenLists((GLuint) last+1);
  w->font_range[isize]=last+1;
  if (w->font_base[isize] == 0) 
    {
      glwError("out of display lists\n");
      exit(1);
    }
  
  glXUseXFont(id, first, last-first+1, w->font_base[isize]+first);
  /*    *height = font_info->ascent + font_info->descent;
   *width = font_info->max_bounds.width;  */
  return 1;
}



void glwPrint(glWindow w,char *s)
{

  if (!glwPrintFB(w,s))
    {
      if (w->font_base[w->font_to_use]==0) return;
      glPushAttrib (GL_LIST_BIT);
      glListBase(w->font_base[w->font_to_use]);
      glCallLists((int)strlen(s), GL_UNSIGNED_BYTE, (GLubyte *) s);
      glPopAttrib ();
    }
}

void glwPrintf(glWindow w,char *format,...)
{ 
  va_list ptr;
  va_start(ptr,format);
  vsprintf(glw_string_buffer,format,ptr);
  va_end(ptr);
  glwPrint(w,glw_string_buffer);
}


void glwSetFontSize(glWindow w, double font_size)
{
  int isize;
  isize=(int)(font_size*(double)w->h);
  if (isize<2) isize=2;
  if (isize>=GLW_MAX_FONT_SIZE) isize=GLW_MAX_FONT_SIZE-1;
  w->font_size=font_size;
  w->font_to_use=isize;
#ifdef PTHREADS
  
  if (glw_threads && pthread_equal(pthread_self(),glw_thread))
    glwLoadFont(w,w->font_to_use);
#endif
}

double glwGetFontSize(glWindow w)
{
  return w->font_size;
}
/******************************************************************************/

void glwGetMouseLoc(glWindow w,int *x, int *y)
{
    int junk;

    *x = 0;
    *y = 0;
    XQueryPointer(glw_display, w->w_main, (Window *)&junk, (Window *)&junk,
		  &junk, &junk, x, y, (unsigned int *)&junk);
}

/******************************************************************************/

void glwSetFogRamp(glWindow w,int density, int startIndex)
{
    XColor c[256];
    int rShift, gShift, bShift, intensity, fogValues, colorValues;
    int i, j, k;

    switch (w->vInfoMainClass) {
      case DirectColor:
	fogValues = 1 << density;
	colorValues = 1 << startIndex;
	for (i = 0; i < colorValues; i++) {
	    for (j = 0; j < fogValues; j++) {
		k = i * fogValues + j;
		intensity = i * fogValues + j * colorValues;
		if (intensity > w->vInfoMain->colormap_size) {
		    intensity = w->vInfoMain->colormap_size;
		}
		intensity = (intensity << 8) | intensity;
		rShift = ffs((unsigned int)w->vInfoMain->red_mask) - 1;
		gShift = ffs((unsigned int)w->vInfoMain->green_mask) - 1;
		bShift = ffs((unsigned int)w->vInfoMain->blue_mask) - 1;
		c[k].pixel = ((k << rShift) & w->vInfoMain->red_mask) |
			     ((k << gShift) & w->vInfoMain->green_mask) |
			     ((k << bShift) & w->vInfoMain->blue_mask);
		c[k].red = (unsigned short)intensity;
		c[k].green = (unsigned short)intensity;
		c[k].blue = (unsigned short)intensity;
		c[k].flags = DoRed | DoGreen | DoBlue;
	    }
	}
	XStoreColors(glw_display, w->cMapMain, c, w->vInfoMain->colormap_size);
	break;
      case GrayScale:
      case PseudoColor:
	fogValues = 1 << density;
	colorValues = 1 << startIndex;
	for (i = 0; i < colorValues; i++) {
	    for (j = 0; j < fogValues; j++) {
		k = i * fogValues + j;
		intensity = i * fogValues + j * colorValues;
		if (intensity > w->vInfoMain->colormap_size) {
		    intensity = w->vInfoMain->colormap_size;
		}
		intensity = (intensity << 8) | intensity;
		c[k].pixel = k;
		c[k].red = (unsigned short)intensity;
		c[k].green = (unsigned short)intensity;
		c[k].blue = (unsigned short)intensity;
		c[k].flags = DoRed | DoGreen | DoBlue;
	    }
	}
	XStoreColors(glw_display, w->cMapMain, c, w->vInfoMain->colormap_size);
	break;
    }

    XSync(glw_display, 0);
}

/******************************************************************************/

void glwSetGreyRamp(glWindow w)
{
    XColor c[256];
    float intensity;
    int rShift, gShift, bShift, i;

    switch (w->vInfoMainClass) {
      case DirectColor:
	for (i = 0; i < w->vInfoMain->colormap_size; i++) {
	    intensity = (float)i / (float)w->vInfoMain->colormap_size *
			65535.0 + 0.5;
	    rShift = ffs((unsigned int)w->vInfoMain->red_mask) - 1;
	    gShift = ffs((unsigned int)w->vInfoMain->green_mask) - 1;
	    bShift = ffs((unsigned int)w->vInfoMain->blue_mask) - 1;
	    c[i].pixel = ((i << rShift) & w->vInfoMain->red_mask) |
			 ((i << gShift) & w->vInfoMain->green_mask) |
			 ((i << bShift) & w->vInfoMain->blue_mask);
	    c[i].red = (unsigned short)intensity;
	    c[i].green = (unsigned short)intensity;
	    c[i].blue = (unsigned short)intensity;
	    c[i].flags = DoRed | DoGreen | DoBlue;
	}
	XStoreColors(glw_display, w->cMapMain, c, w->vInfoMain->colormap_size);
	break;
      case GrayScale:
      case PseudoColor:
	for (i = 0; i < w->vInfoMain->colormap_size; i++) {
	    intensity = (float)i / (float)w->vInfoMain->colormap_size *
			65535.0 + 0.5;
	    c[i].pixel = i;
	    c[i].red = (unsigned short)intensity;
	    c[i].green = (unsigned short)intensity;
	    c[i].blue = (unsigned short)intensity;
	    c[i].flags = DoRed | DoGreen | DoBlue;
	}
	XStoreColors(glw_display, w->cMapMain, c, w->vInfoMain->colormap_size);
	break;
    }

    XSync(glw_display, 0);
}

/******************************************************************************/

void glwSetOneColor(glWindow w,int index, float r, float g, float b)
{
    XColor c;
    int rShift, gShift, bShift;

    switch (w->vInfoMainClass) {
      case DirectColor:
	rShift = ffs((unsigned int)w->vInfoMain->red_mask) - 1;
	gShift = ffs((unsigned int)w->vInfoMain->green_mask) - 1;
	bShift = ffs((unsigned int)w->vInfoMain->blue_mask) - 1;
	c.pixel = ((index << rShift) & w->vInfoMain->red_mask) |
		  ((index << gShift) & w->vInfoMain->green_mask) |
		  ((index << bShift) & w->vInfoMain->blue_mask);
	c.red = (unsigned short)(r * 65535.0 + 0.5);
	c.green = (unsigned short)(g * 65535.0 + 0.5);
	c.blue = (unsigned short)(b * 65535.0 + 0.5);
	c.flags = DoRed | DoGreen | DoBlue;
	XStoreColor(glw_display, w->cMapMain, &c);
	break;
      case GrayScale:
      case PseudoColor:
	if (index < w->vInfoMain->colormap_size) {
	    c.pixel = index;
	    c.red = (unsigned short)(r * 65535.0 + 0.5);
	    c.green = (unsigned short)(g * 65535.0 + 0.5);
	    c.blue = (unsigned short)(b * 65535.0 + 0.5);
	    c.flags = DoRed | DoGreen | DoBlue;
	    XStoreColor(glw_display, w->cMapMain, &c);
	}
	break;
    }

    XSync(glw_display, 0);
}

/******************************************************************************/

void glwSetOverlayMap(glWindow w, int size, float *rgb)
{
    XColor c;
    unsigned long *buf;
    int max, i;

    if (w->vInfoMainClass == PseudoColor) {
	max = (size > w->vInfoOverlay->colormap_size) ?
	      w->vInfoOverlay->colormap_size : size;
	buf = (unsigned long *)calloc(max, sizeof(unsigned long));
	XAllocColorCells(glw_display, w->cMapOverlay, True, NULL, 0, buf, max-1);
	for (i = 1; i < max; i++) {
	    c.pixel = i;
	    c.red = (unsigned short)(rgb[i] * 65535.0 + 0.5);
	    c.green = (unsigned short)(rgb[size+i] * 65535.0 + 0.5);
	    c.blue = (unsigned short)(rgb[size*2+i] * 65535.0 + 0.5);
	    c.flags = DoRed | DoGreen | DoBlue;
	    XStoreColor(glw_display, w->cMapOverlay, &c);
	}
	free(buf);
    }

    XSync(glw_display, 0);
}


/******************************************************************************/

void glwSetRGBMap(glWindow w,int size, float *rgb)
{
    XColor c;
    int rShift, gShift, bShift, max, i;

    switch (w->vInfoMainClass) {
      case DirectColor:
	max = (size > w->vInfoMain->colormap_size) ? w->vInfoMain->colormap_size
						  : size;
	for (i = 0; i < max; i++) {
	    rShift = ffs((unsigned int)w->vInfoMain->red_mask) - 1;
	    gShift = ffs((unsigned int)w->vInfoMain->green_mask) - 1;
	    bShift = ffs((unsigned int)w->vInfoMain->blue_mask) - 1;
	    c.pixel = ((i << rShift) & w->vInfoMain->red_mask) |
		      ((i << gShift) & w->vInfoMain->green_mask) |
		      ((i << bShift) & w->vInfoMain->blue_mask);
	    c.red = (unsigned short)(rgb[i] * 65535.0 + 0.5);
	    c.green = (unsigned short)(rgb[size+i] * 65535.0 + 0.5);
	    c.blue = (unsigned short)(rgb[size*2+i] * 65535.0 + 0.5);
	    c.flags = DoRed | DoGreen | DoBlue;
	    XStoreColor(glw_display, w->cMapMain, &c);
	}
	break;
      case GrayScale:
      case PseudoColor:
	max = (size > w->vInfoMain->colormap_size) ? w->vInfoMain->colormap_size
						  : size;
	for (i = 0; i < max; i++) {
	    c.pixel = i;
	    c.red = (unsigned short)(rgb[i] * 65535.0 + 0.5);
	    c.green = (unsigned short)(rgb[size+i] * 65535.0 + 0.5);
	    c.blue = (unsigned short)(rgb[size*2+i] * 65535.0 + 0.5);
	    c.flags = DoRed | DoGreen | DoBlue;
	    XStoreColor(glw_display, w->cMapMain, &c);
	}
	break;
    }

    XSync(glw_display, 0);
}



/******************************************************************************/

GLenum glwSetWindowLevel(glWindow w,GLenum level)
{

    switch (level) {
      case GLW_OVERLAY:
	if (GLW_HAS_OVERLAY(w->type)) {
	    if (!glXMakeCurrent(glw_display, w->w_overlay, w->cOverlay)) {
		return GL_FALSE;
	    }
	} else {
	    return GL_FALSE;
	}
	break;
      case GLW_RGB:
      case GLW_INDEX:
	if (!glXMakeCurrent(glw_display, w->w_main, w->cMain)) {
	    return GL_FALSE;
	}
	break;
    default:
	break;
    }
    return GL_TRUE;
}

void glwQuit(glWindow w)
{
  glw_process_events=0;
}
/******************************************************************************/


static void glw_next_event()
{
  glWindow w;
  Window xw;
  XEvent current, ahead;
  char buf[1000];
  KeySym ks;
  int key;
  GLenum redraw_request;
  glwTrace(0,"nextev");
  XNextEvent(glw_display, &current);
  glwTrace(0,"recev");
  xw=current.xany.window;
  redraw_request= GL_FALSE;
  switch (current.type) 
    {
      /* events where interesting window id is in a different place*/
    case CirculateNotify:
    case CirculateRequest:
    case ConfigureNotify:
    case ConfigureRequest:
    case CreateNotify:
    case DestroyNotify:
    case GravityNotify:
    case MapNotify:
    case MapRequest:
    case ReparentNotify:
    case UnmapNotify:
      xw = current.xmaprequest.window;
      break;
    }
  
  w=glw_find(xw);
  /* window might have been destroyed */
  if (w==NULL) 
    {
      return;
    }
      
  assert(w->w_main==xw);
  switch (current.type) 
    {
    case MappingNotify:
      XRefreshKeyboardMapping((XMappingEvent *)&current);
      w->lastEventType = MappingNotify;
      break;

    case MapNotify:
      w->lastEventType = MapNotify;
      w->drawAllowFlag = GL_TRUE;
      break;
      
    case UnmapNotify:
      w->lastEventType = UnmapNotify;
      w->drawAllowFlag = GL_FALSE;
      break;

    case ClientMessage:
      w->lastEventType = ClientMessage;
      if (current.xclient.data.l[0] == w->deleteWindowAtom) 
	{
	  exit(0);
	}
      else if (current.xclient.data.l[1] == 123) 
	{
	  fprintf(stderr,"got it\n");
	  redraw_request= GL_TRUE;
	}
      break;
      
    case Expose:
      while (XEventsQueued(current.xexpose.display, QueuedAfterReading) > 0) {
	XPeekEvent(current.xexpose.display, &ahead);
	if (ahead.xexpose.window != current.xexpose.window ||
	    ahead.type != Expose) {
	  break;
	}
	XNextEvent(glw_display, &current);
      }

      if (current.xexpose.count == 0) 
	{
	  if (w->expose_func) {
	    w->drawAllowFlag = GL_TRUE;
	    w->lastEventType = Expose;
	    (*w->expose_func)(w,w->info,w->w, w->h);
	    if (w->lastEventType != ConfigureNotify) 
	      redraw_request= GL_TRUE;
	  }
	}
      break;
      
    case ConfigureNotify:
      w->lastEventType = ConfigureNotify;
      w->w = current.xconfigure.width;
      w->h = current.xconfigure.height;
      glwSetFontSize(w, w->font_size);
      if (w->reshape_func) 
	{
	  (*w->reshape_func)(w,w->info,w->w, w->h);
	  redraw_request= GL_TRUE;
	}
      break;
    case MotionNotify:
      w->lastEventType = MotionNotify;
      if (w->mouse_move_func) {
	GLenum mask;
	
	mask = (GLenum)0;
	if (current.xmotion.state & Button1Mask)  mask |= GLW_LEFTBUTTON;
	if (current.xmotion.state & Button2Mask)  mask |= GLW_MIDDLEBUTTON;
	if (current.xmotion.state & Button3Mask)  mask |= GLW_RIGHTBUTTON;
	if (current.xmotion.state & ShiftMask)    mask |= GLW_SHIFT;
	if (current.xmotion.state & ControlMask)  mask |= GLW_CONTROL;
	if (current.xmotion.state & Mod1Mask)     mask |= GLW_MOD1;
	if (current.xkey.state & LockMask)        mask |= GLW_LOCK;
	    
	redraw_request= (*w->mouse_move_func)(w,w->info,current.xmotion.x, current.xmotion.y, mask);
       }
      break;
    case ButtonPress:
      w->lastEventType = ButtonPress;
      if (w->mouse_down_func) 
	{
	  GLenum mask;
	  
	  mask = (GLenum)0;
	  if (current.xbutton.button==1)  mask |= GLW_LEFTBUTTON;
	  if (current.xbutton.button==2)  mask |= GLW_MIDDLEBUTTON;
	  if (current.xbutton.button==3)  mask |= GLW_RIGHTBUTTON;
	  if (current.xbutton.state & ShiftMask)    mask |= GLW_SHIFT;
	  if (current.xbutton.state & ControlMask)  mask |= GLW_CONTROL;
	  if (current.xmotion.state & Mod1Mask)     mask |= GLW_MOD1;
	  
	redraw_request= (*w->mouse_down_func)(w,w->info,current.xbutton.x, current.xbutton.y, mask);
      }
      break;
      
    case ButtonRelease:
        w->lastEventType = ButtonRelease;
        if (w->mouse_up_func) {
	  GLenum mask;
	  mask = (GLenum) 0;
	  if (current.xbutton.button==1)  mask |= GLW_LEFTBUTTON;
	  if (current.xbutton.button==2)  mask |= GLW_MIDDLEBUTTON;
	  if (current.xbutton.button==3)  mask |= GLW_RIGHTBUTTON;
	  if (current.xbutton.state & ShiftMask)    mask |= GLW_SHIFT;
	  if (current.xbutton.state & ControlMask)  mask |= GLW_CONTROL;
	  if (current.xmotion.state & Mod1Mask)     mask |= GLW_MOD1;
	  
	  redraw_request= (*w->mouse_up_func)(w,w->info,current.xbutton.x, current.xbutton.y, mask);
	  break;
         }
	break;
      case KeyPress:
        w->lastEventType = KeyPress;
        XLookupString(&current.xkey, buf, sizeof(buf), &ks, 0);
        switch (ks) {
#define CaseKS(sym) case XK_##sym: key=GLW_##sym; break
	  CaseKS(0);
          CaseKS(1);
          CaseKS(2);
          CaseKS(3);
          CaseKS(4);
          CaseKS(5);
          CaseKS(6);
          CaseKS(7);
          CaseKS(8);
          CaseKS(9);
          CaseKS(A);
          CaseKS(B);
          CaseKS(C);
          CaseKS(D);
          CaseKS(E);
          CaseKS(F);
          CaseKS(G);
          CaseKS(H);
          CaseKS(I);
          CaseKS(J);
          CaseKS(K);
          CaseKS(L);
          CaseKS(M);
          CaseKS(N);
          CaseKS(O);
          CaseKS(P);
          CaseKS(Q);
          CaseKS(R);
          CaseKS(S);
          CaseKS(T);
          CaseKS(U);
          CaseKS(V);
          CaseKS(W);
          CaseKS(X);
          CaseKS(Y);
          CaseKS(Z);
          CaseKS(a);
          CaseKS(b);
          CaseKS(c);
          CaseKS(d);
          CaseKS(e);
          CaseKS(f);
          CaseKS(g);
          CaseKS(h);
          CaseKS(i);
          CaseKS(j);
          CaseKS(k);
          CaseKS(l);
          CaseKS(m);
          CaseKS(n);
          CaseKS(o);
          CaseKS(p);
          CaseKS(q);
          CaseKS(r);
          CaseKS(s);
          CaseKS(t);
          CaseKS(u);
          CaseKS(v);
          CaseKS(w);
          CaseKS(x);
          CaseKS(y);
          CaseKS(z);
	  CaseKS(space);
	  CaseKS(exclam);
	  CaseKS(quotedbl);
	  CaseKS(numbersign);
	  CaseKS(dollar);
	  CaseKS(percent);
	  CaseKS(ampersand);
	  CaseKS(apostrophe);
	  CaseKS(parenleft);
	  CaseKS(parenright);
	  CaseKS(asterisk);
	  CaseKS(plus);
	  CaseKS(comma);
	  CaseKS(minus);
	  CaseKS(period);
	  CaseKS(slash);
	  CaseKS(colon);
	  CaseKS(semicolon);
	  CaseKS(less);
	  CaseKS(equal);
	  CaseKS(greater);
	  CaseKS(question);
	  CaseKS(at);
	  CaseKS(bracketleft);
	  CaseKS(bracketright);
	  CaseKS(asciicircum);
	  CaseKS(underscore);
	  CaseKS(grave);
	  CaseKS(braceleft);
	  CaseKS(bar);
	  CaseKS(braceright);
	  CaseKS(asciitilde);
	  CaseKS(Return);
	  CaseKS(BackSpace);
	  CaseKS(Escape);
	  CaseKS(Left);
	  CaseKS(Up);
	  CaseKS(Right);
	  CaseKS(Down);
	  CaseKS(KP_Enter);
	  CaseKS(KP_Home);
	  CaseKS(KP_Left);
	  CaseKS(KP_Up);
	  CaseKS(KP_Right);
	  CaseKS(KP_Down);
	  CaseKS(KP_Page_Up);
	  CaseKS(KP_Page_Down);
	  CaseKS(KP_End);
	  CaseKS(KP_Begin);
	  CaseKS(KP_Insert);
	  CaseKS(KP_Delete);
	  CaseKS(KP_Divide);
	  CaseKS(KP_Multiply);
	  CaseKS(KP_Add);
	  CaseKS(KP_Subtract);
	  CaseKS(KP_Decimal);
	  CaseKS(KP_0);
	  CaseKS(KP_1);
	  CaseKS(KP_2);
	  CaseKS(KP_3);
	  CaseKS(KP_4);
	  CaseKS(KP_5);
	  CaseKS(KP_6);
	  CaseKS(KP_7);
	  CaseKS(KP_8);
	  CaseKS(KP_9);
	  CaseKS(F1);
	  CaseKS(F2);
	  CaseKS(F3);
	  CaseKS(F4);
	  CaseKS(F5);
	  CaseKS(F6);
	  CaseKS(F7);
	  CaseKS(F8);
	  CaseKS(F9);
	  CaseKS(F10);
	  CaseKS(F11);
	  CaseKS(F12);
	  CaseKS(Page_Up);
	  CaseKS(Page_Down);
	  CaseKS(Tab);
	  CaseKS(Home);
	  CaseKS(End);
	  CaseKS(Insert);
	  CaseKS(Delete);
	  CaseKS(Pause);
	  CaseKS(Scroll_Lock);
	  CaseKS(Print);
	default:              key = GL_FALSE;         break;
        }
	if (key && w->key_down_func)
	  {
	    GLenum mask;
	    
	    mask = (GLenum)0;
	    if (current.xkey.state & ShiftMask)    mask |= GLW_SHIFT;
	    if (current.xkey.state & ControlMask)  mask |= GLW_CONTROL;
	    if (current.xkey.state & Mod1Mask)     mask |= GLW_MOD1;
	    
	    redraw_request= (*w->key_down_func)(w,w->info,key, mask);
	  }
	break;
    }
  w->event_redraw_request=redraw_request;
}




void glwSwapBuffers(glWindow w)
{
  if (w->swap) return;
  w->swap=1;
  if (!w->offscreen)
    glXSwapBuffers(glw_display, w->w_main);
    
  if (!w->firstswap) glwDumpOnSwap(w);
  w->firstswap=0;
  w->swap=0;

}

int glwAttach(glWindow w)
{
  if (glw_current==w) return 1; 
  if (w==NULL) { glwError("null window"); return GL_FALSE;}
  if (w->offscreen)
    {
	if (!glXMakeCurrent(glw_display, w->glx_pixmap, w->cMain)) 
	  {
	    return 0;
	  }
    }
  else
    {
      if (!glXMakeCurrent(glw_display, w->w_main, w->cMain)) 
	{
	  return 0;
	}
    }

  glw_current=w;
  return 1;
}

void glwFlush(glWindow w)
{
  glwTrace(w,"flush");
  glXWaitGL();
  XFlush(glw_display);
}



/*
http://www.faqs.org/faqs/x-faq/part7/:

Subject: 161)  How do I simulate a button press/release event for a widget?

You can do this using XSendEvent(); it's likely that you're not setting the
window field in the event, which Xt needs in order to match to the widget
which should receive the event.

If you're sending events to your own application, then you can use
XtDispatchEvent() instead. This is more efficient than XSendEvent() in that
you avoid a round-trip to the server.

Depending on how well the widget was written, you may be able to call its
action procedures in order to get the effects you want.

[courtesy Mark A. Horstman (mh2620@sarek.sbc.com), 11/90]



Subject: 163)! What are these "Xlib: unexpected async reply" errors?

You'll typically get these errors if you are writing a multi-threaded
application and are making X calls from more than one thread -- one of the
more common new ways to introduce memory corruption into Xlib (using bogus
pointers is another, as is mixing up XFree/XtFree/free calls. Even an
operation as simple as XSendEvent can't be called from a second thread.).
Prior to X11R6, X doesn't support multi-threading; check the X11R6
documentation for how to write a threaded application safely with X11R6 and
later versions of X (including being sure to enable Xlib's multi-thread
support).

Common widget sets are also unsafe to use except from within one thread.  The
versions of Motif from OSF, for example, offer no support for
multi-threading.

*/

static void glw_sleep_ms(unsigned long ms)
{
  struct timespec t;
  t.tv_sec=0;
  t.tv_nsec=ms*1000000;
  /* 10 ms */
  nanosleep(&t,NULL);
}


static void glw_main_loop(void);
void glwProcess(glWindow w,void *info)
{
  w->info=info;
#ifdef PTHREADS
  { 
    if(glw_threads && w->application_redraw_request)
      {
	pthread_mutex_lock(&glw_precond_mutex);
	pthread_cond_wait(&glw_precond,&glw_precond_mutex);
	pthread_mutex_unlock(&glw_precond_mutex);
      }
  }
#endif
  if (glw_control_mode==GLW_EVENT_DRIVEN) glw_process_events=1;
  glwLock();
  w->application_redraw_request=1;
  glwUnlock();
#ifdef PTHREADS
  if (glw_threads && glw_control_mode==GLW_EVENT_DRIVEN)
    {
      pthread_mutex_lock(&glw_cond_mutex);
      pthread_cond_wait(&glw_cond,&glw_cond_mutex);
      pthread_mutex_unlock(&glw_cond_mutex);
    }
#endif
  if (!glw_threads)
    glw_main_loop();
}


/******************************************************************************/

void glwSetRedrawFunc(glWindow w, void (*Func)(glWindow, void *))
{
    if (w==NULL) { glwError("null window"); return;}
    w->redraw_func = Func;
}


void glwSetControlMode(glWindow w, int mode)
{
  glw_control_mode=mode;
}
/******************************************************************************/

void glwSetExposeFunc(glWindow w, void (*Func)(glWindow, void *,int, int))
{    
    if (w==NULL) { glwError("null window"); return;}
    w->expose_func = Func;
}

/******************************************************************************/

void glwSetReshapeFunc(glWindow w,void (*Func)(glWindow, void *,int, int))
{
    if (w==NULL) { glwError("null window"); return;}

    w->reshape_func = Func;
}

/******************************************************************************/

void glwSetDisplayFunc(glWindow w, void (*Func)(glWindow, void *))
{
  if (w==NULL) { glwError("null window"); return;}
  if (Func!=NULL)
    {
      w->redraw_func = Func;
      glw_control_mode=GLW_EVENT_DRIVEN;
    }
  else
    glw_control_mode=GLW_APPLICATION_DRIVEN;
  
}
/******************************************************************************/

void glwSetKeyDownFunc(glWindow w,GLenum (*Func)(glWindow, void *,int, GLenum))
{
    if (w==NULL) { glwError("null window"); return;}

    w->key_down_func = Func;
}

/******************************************************************************/

void glwSetMouseDownFunc(glWindow w,GLenum (*Func)(glWindow, void *,int, int, GLenum))
{

   if (w==NULL) { glwError("null window"); return;}
   w->mouse_down_func = Func;
}

/******************************************************************************/

void glwSetMouseUpFunc(glWindow w,GLenum (*Func)(glWindow, void *,int, int, GLenum))
{
    if (w==NULL) { glwError("null window"); return;}

    w->mouse_up_func = Func;
}


/******************************************************************************/

void glwSetMouseMoveFunc(glWindow w,GLenum (*Func)(glWindow, void *,int, int, GLenum))
{
    if (w==NULL) { glwError("null window"); return;}

    w->mouse_move_func = Func;
}

/******************************************************************************/

void glwSetIdleFunc(glWindow w,void (*Func)(glWindow, void *))
{
  if (w==NULL) { glwError("null window"); return;}
  if (Func!=NULL)
    {
      w->redraw_func = Func;
      glw_control_mode=GLW_APPLICATION_DRIVEN;
    }
  else
    glw_control_mode=GLW_EVENT_DRIVEN;
}

/******************************************************************************/

void glw_destroy(glWindow w)
{
   int i;
   glwTrace(w,"destroy:");
   XFlush(glw_display);
   /*   XLockDisplay(glw_display);*/
   if (glwAttach(w)) 
     {   
   for (i=0;i<GLW_MAX_FONT_SIZE;i++)
     {
       if (w->font_base[i]!=0) glDeleteLists(w->font_base[i],w->font_range[i]);
     }
       glFlush();
       glFinish();
     }
   
   glw_remove(w);
  glXMakeCurrent(glw_display, None,NULL);
   
   if (GLW_HAS_OVERLAY(w->type)) 
     {
       glXDestroyContext(glw_display, w->cOverlay);
       XDestroyWindow(glw_display, w->w_overlay);
       XFreeColormap(glw_display, w->cMapOverlay);
       XFree((char *)w->vInfoOverlay);
       glwTrace(w,"destroy overlay");
       
     }
   if (w->offscreen)
     {
       glXDestroyGLXPixmap(glw_display, w->glx_pixmap);
       XFreePixmap(glw_display, w->pixmap);
       glwTrace(w,"destroy pixmap");
     }
   else
     {
       glXDestroyContext(glw_display, w->cMain);
       glwTrace(w,"destroy xwindow");
     }
   XFreeColormap(glw_display, w->cMapMain);
   XFree((char *)w->vInfoMain);
   XDestroyWindow(glw_display, w->w_main);
   if (w->title!=NULL) free(w->title);
   if (w->xtitle!=NULL) free(w->xtitle);
   if (w->state!=NULL) free(w->state);
   /*   free(w);*/
   w=NULL;
   glwTrace(w,"destroy window");
   
   /*   XUnlockDisplay(glw_display);*/
}

void glwDestroy(glWindow w)
{
  w->destroy_request=1;
  if (!glw_threads)
    glw_main_loop();

}

void glwFinish(void)
{
  int i;
  XCloseDisplay(glw_display);
  for (i=0;i<GLW_MAX_FONT_SIZE;i++)
    {
      if (glw_font_info[i]!=0) XFreeFont(glw_display,glw_font_info[i]);
    }
}
/******************************************************************************/
char *glwGetDefault(char *resource, char *dflt)
{
  char *data;
  data=XGetDefault(glw_display,"gltools",resource);
  if (data==NULL) 
    return dflt;
  else
    return data;
}

/******************************************************************************/
void glwGetWindowSize(glWindow w, int *width, int *height)
{
  *width=w->w;
  *height=w->h;
}
/******************************************************************************/
int glwIsMesa(glWindow w)
{
  return 0;
}

/******************************************************************************/
static int  glw_default_x;
static int  glw_default_y;
static int  glw_default_w;
static int  glw_default_h;

static glWindow glw_xcreate(glWindow w);
void glw_process_window(glWindow w)
{

  GLenum need_redraw;
  glwTrace(w,"process");
  
  if (w->create_request) 
    {
      glwTrace(w,"xcreate");
      /*      XLockDisplay(glw_display);*/
      w=glw_xcreate(w);
      /*      XUnlockDisplay(glw_display);*/
      if (w==NULL) 
	{
	  glwError("Failed to open window");
	  exit(1);
	}
    }
  if (!glwAttach(w)) 
    {
      glwTrace(w,"wait4attach")
      glXWaitGL();
      w->lastEventType= -1;
      return;
    }
  if (w->font_base[w->font_to_use]==0) glwLoadFont(w,w->font_to_use);

  glwTrace(w,"need_redraw");

  if (w->destroy_request)
    {

      glw_destroy(w);
      return ;
    }
  glwLock();
  need_redraw=w->application_redraw_request||w->event_redraw_request;
  glwUnlock();

  glwTrace(w,"need_redraw");

  if ((w->redraw_func)&&need_redraw&&w->drawAllowFlag)
    {
      XDefineCursor(glw_display, w->w_main,w->wait_cursor);
      glwTrace(w,"redraw");
      (*w->redraw_func)(w,w->info);
      glXWaitGL();
      need_redraw = 0;
      glwLock();
      w->application_redraw_request=0;
      glwUnlock();
      w->event_redraw_request=0;
      XDefineCursor(glw_display, w->w_main,w->user_cursor);
    }
  w->lastEventType= -1;
  glwTrace(w,"glw_process_window");
}

static void   glw_init_x(void);

static void glw_main_loop(void)
{
  struct timespec t;
  static int control_mode;
  static int redraw_request;
  glWindow w;
#ifdef PTHREADS
  glwTrace(0,"init X");
  glw_init_x();
  glwTrace(0,"init X");
#endif
  glwTrace(0,"main_loop");
  do
    {
      glwTrace(0,"loop");
      control_mode=glw_control_mode;
      /*      if (glw_process_events)
	      glw_next_event();*/

      while (XPending(glw_display)) 
	{
	  glw_next_event();
	}

      glwLock();
      redraw_request=0;
      for (w=glw_anchor;w&& !redraw_request;w=w->next)
	{
	  redraw_request=
	    w->create_request||
	    w->application_redraw_request||
	    w->event_redraw_request;
	}
      glwUnlock();
#ifdef PTHREADS
      if (!glw_process_events&& !redraw_request)
	{
	      glw_sleep_ms(1);
	}
#endif

      for (w=glw_anchor;w;w=w->next)
	{
	  glw_process_window(w);
	}
#ifdef PTHREADS
      if (glw_threads)
	{
	  pthread_cond_broadcast(&glw_precond); /*???*/
	  if (control_mode==GLW_EVENT_DRIVEN&& glw_control_mode==GLW_APPLICATION_DRIVEN)
	    pthread_cond_broadcast(&glw_cond);
	}
#endif
      XFlush(glw_display);

    } while (glw_process_events || glw_threads);

}

static void glw_init_x(void)
{

  int erb, evb,i;
      glwTrace(0,"init X");
#ifdef PTHREADS
  /* XInitThreads();*/
#endif  
  if (glw_display!=NULL) 
    {
      glwTrace(0,"init already done");
      return;
    }
  glw_display=XOpenDisplay(NULL);
  if (glw_display==NULL) 
    {
      glwTrace(0,"init error");
      glwError("Unable to connect to display!");
      exit(1);
    }
  /*  XSetErrorHandler(ErrorHandler);*/
  glw_screen=DefaultScreen(glw_display);
  glw_root=RootWindow(glw_display, glw_screen);
  if (!glXQueryExtension(glw_display, &erb, &evb))
    {
      glwError("OpenGL is not installed for your display!");
      glwTrace(0,"init error");
      exit(1);
    }

  for(i=0;i<GLW_MAX_FONT_SIZE;i++)  glw_font_info[i]=0;
  glw_default_x=atoi(glwGetDefault("xOrigin","0"));
  glw_default_y=atoi(glwGetDefault("yOrigin","0"));;
  glw_default_w=(atoi(glwGetDefault("width","352"))/16)*16; /* GLW_MPEG1_PAL */
  glw_default_h=(atoi(glwGetDefault("height","288"))/16)*16; /* GLW_MPEG1_PAL */
  glwTrace(0,"initx ok");

}

static int initialized=0;
static void glw_init_module(void)
{
  if (initialized) return;
  initialized=1;
  glw_anchor=NULL;
  if (glw_threads==-1)
    {
#ifdef PTHREADS
      glw_threads=1;
#else
      glw_threads=0;
#endif
    }
#ifdef PTHREADS
  if (glw_threads)
    {
      pthread_mutex_init(&glw_mutex,0);  
      pthread_mutex_init(&glw_cond_mutex,0);  
      pthread_cond_init(&glw_cond,0);  
      pthread_mutex_init(&glw_precond_mutex,0);  
      pthread_cond_init(&glw_precond,0);  
      pthread_attr_init(&glw_attr);  
      pthread_attr_setstacksize(&glw_attr,265*1000);
      assert(pthread_create(&glw_thread,&glw_attr,(void*(*)(void*)) glw_main_loop, 0)==0);
      glwTrace(0,"thread ok");
    }
#endif      
  if (!glw_threads)
    glw_init_x();
}

static glWindow glw_init=0;
static void glw_init_reset(void)
{
  glw_init=0;
}
static glWindow glw_init_window(void)
{
  glWindow w;
  glw_init_module();
  if (glw_init) return glw_init;
  
  w=(struct glWindowStruct *)calloc(sizeof(struct glWindowStruct),1);
  if (w==NULL) 
    {
      glwError("Failed to allocate memory for window struct.");
      exit(1);
    }
  
  w->x=glw_default_x;
  w->y=glw_default_y;
  w->w=glw_default_w;
  w->h=glw_default_h;
  w->type=(GLenum)(GLW_RGB|GLW_DOUBLE|GLW_DIRECT|GLW_DEPTH);
  w->dm_policy=(GLenum)GLW_MINIMUM_CRITERIA;
  w->title=(char*)malloc(GLW_NAMELEN);
  w->xtitle=(char*)malloc(GLW_NAMELEN);
  w->state=NULL;
  /*  strcpy(w->title,"glwGetDefault("title","gltools"));*/
  strcpy(w->title,"gltools");
  glw_control_mode=GLW_EVENT_DRIVEN;
  w->font_base[0]=0;
  w->keep_aspect=1;
  w->offscreen=0;
  w->w_toplevel=0;
  w->swap=0;
  w->firstswap=1;
  w->create_request=1;
  w->destroy_request=0;
  glwInitFB(w);
  glw_init=w;
  return w;
}

void glwInitDisplayMode(GLenum type)
{   
  glWindow w=glw_init_window();
  w->type = type;
}

void glwInitOffscreen(void)
{   
  glWindow w=glw_init_window();
  w->offscreen=1;
  glw_control_mode=GLW_APPLICATION_DRIVEN;
}

void glwInitDisplayModePolicy(GLenum type)
{
  glWindow w=glw_init_window();
  w->dm_policy = type;
}

void glwInitToplevel(void *toplevel)
{
  glWindow w=glw_init_window();
  w->w_toplevel=(Window)toplevel;
}



void glwInitPosition(int x, int y, int width, int height)
{
  
  glWindow w=glw_init_window();
  w->x = x;
  w->y = y;
  w->w = width;
  w->h = height;
  /*  w->w = (width/16)*16;
  w->h = (height/16)*16;*/
}

void glwInitPos(int x, int y)
{
  glWindow w=glw_init_window();
  w->x = x;
  w->y = y;
}


void glwInitTitle(char *title)
{
  glWindow w=glw_init_window();
  strcpy(w->title,title);
}


void glwInitAspectKeeping(int yesno)
{
  glWindow w=glw_init_window();
  w->keep_aspect=yesno;
}

 

static XVisualInfo *FindBestMainVisual(glWindow w,GLenum type)
{
    int list[128], i;
    XVisualInfo *vinfo;
    glwTrace(w,"FindBestMainVisual");
    i = 0;

    list[i++] = GLX_LEVEL;
    list[i++] = 0;

    if (GLW_IS_DOUBLE(type)) {
	list[i++] = GLX_DOUBLEBUFFER;
    }

    if (GLW_IS_RGB(type)) {
	list[i++] = GLX_RGBA;
	list[i++] = GLX_RED_SIZE;
	list[i++] = 1;
	list[i++] = GLX_GREEN_SIZE;
	list[i++] = 1;
	list[i++] = GLX_BLUE_SIZE;
	list[i++] = 1;
	if (GLW_HAS_ALPHA(type)) {
	    list[i++] = GLX_ALPHA_SIZE;
	    list[i++] = 1;
	}
	if (GLW_HAS_ACCUM(type)) {
	    list[i++] = GLX_ACCUM_RED_SIZE;
	    list[i++] = 1;
	    list[i++] = GLX_ACCUM_GREEN_SIZE;
	    list[i++] = 1;
	    list[i++] = GLX_ACCUM_BLUE_SIZE;
	    list[i++] = 1;
	    if (GLW_HAS_ALPHA(type)) {
		list[i++] = GLX_ACCUM_ALPHA_SIZE;
		list[i++] = 1;
	    }
	}
    } else if (GLW_IS_INDEX(type)) {
	list[i++] = GLX_BUFFER_SIZE;
	list[i++] = 1;
    }

    if (GLW_HAS_DEPTH(type)) {
	list[i++] = GLX_DEPTH_SIZE;
	list[i++] = 1;
    }

    if (GLW_HAS_STENCIL(type)) {
	list[i++] = GLX_STENCIL_SIZE;
	list[i++] = 1;
    }

    list[i] = (int)None;

    glwTrace(w,"FindBestMainVisual");
    vinfo= glXChooseVisual(glw_display, glw_screen, list);
    glwTrace(w,"FindBestMainVisual");
    return vinfo;
}

static XVisualInfo *FindExactMainVisual(glWindow w, GLenum type)
{
  int i, nvis, val, rval, gval, bval, aval;
  XVisualInfo *vis_list, *this_vis, *best_vis, sampleVis;
  int this_score, best_score;
  
  /* Get list of visuals for this screen */
  sampleVis.screen = glw_screen;
  vis_list = XGetVisualInfo(glw_display, VisualScreenMask, &sampleVis, &nvis);
  
  /* 
   * Loop through the visuals; find first one that matches the attr 
   * specified in type
   */
  best_score = -1; best_vis = NULL;
  for ( i = 0; i < nvis; i++ ) 
    {
      this_vis = &vis_list[i];
      if (debug) 
	fprintf(stderr,"glwin: testing visual id 0x%x: ", (unsigned)this_vis->visualid);
      /* Visual must be supported by GLX */
      if ( glXGetConfig(glw_display, this_vis, GLX_USE_GL, &val) ) continue;
      if ( !val ) 
	{  
	  if (debug) fprintf(stderr,"no GLX support.\n");
	  continue;
	}
      /* Visual must be in main planes which is level 0 */
      glXGetConfig(glw_display, this_vis, GLX_LEVEL, &val);
      if ( val != 0 ) 
	{  
	  if (debug) fprintf(stderr,"no level 0 visual.\n");
	  continue;
	}
      
      /* Color Index or RGBA? It must match the requested value */
      glXGetConfig(glw_display, this_vis, GLX_RGBA, &val);
      if ( GLW_IS_RGB(type) && !val ) 
	{  
	  if (debug) fprintf(stderr,"no RGB support.\n");
	  continue;
	}
      
      if ( GLW_IS_INDEX(type) && val )
	{  
	  if (debug) fprintf(stderr,"no color index support.\n");
	  continue;
	}
      
      
      /* Double buffered or Single buffered? */
      glXGetConfig(glw_display, this_vis, GLX_DOUBLEBUFFER, &val);
      if ( GLW_IS_DOUBLE(type) && !val )
	{  
	  if (debug) fprintf(stderr,"no double buffer support.\n");
	  continue;
	}
      if ( GLW_IS_SINGLE(type) && val ) 
	{  
	  if (debug) fprintf(stderr,"no single buffer support.\n");
	  continue;
	}

      
      /* If accum requested then accum rgb size must be > 0 */
      /* If alpha requested then alpha size must be > 0 */
      /* if accum & alpha requested then accum alpha size must be > 0 */
      if ( GLW_IS_RGB(type) ) 
	{
	  glXGetConfig(glw_display, this_vis, GLX_ACCUM_RED_SIZE, &rval);
	  glXGetConfig(glw_display, this_vis, GLX_ACCUM_GREEN_SIZE, &gval);
	  glXGetConfig(glw_display, this_vis, GLX_ACCUM_BLUE_SIZE, &bval);
	 
	  glXGetConfig(glw_display, this_vis, GLX_ACCUM_ALPHA_SIZE, &aval);
	  if ( GLW_HAS_ACCUM(type) ) 
	    {
	      if ( rval <= 0 || gval <= 0 || bval <= 0 ) 
		{  
		  if (debug) fprintf(stderr,"accum RGB size <0.\n");
		  continue;
		}
	    } 
	  else 
	    {
	      if ( rval > 0 || gval > 0 || bval > 0 || aval > 0 )
		{  
		  if (debug) fprintf(stderr,"accum support not demanded.\n");
		  continue;
		}
	    }
	
	  glXGetConfig(glw_display, this_vis, GLX_ALPHA_SIZE, &val);
	  if ( GLW_HAS_ALPHA(type) ) 
	    {
	      if ( val <= 0 )
		{  
		  if (debug) fprintf(stderr,"alpha size < 0.\n");
		  continue;
		}

	    if ( GLW_HAS_ACCUM(type) && aval <= 0 ) 
	      {
		if (debug) fprintf(stderr,"alpha accum size < 0.\n");
		continue;
	      }
	    }
	  else
	    {
	      if ( val > 0 ) 
	      {
		if (debug) fprintf(stderr,"alpha support not demanded.\n");
		continue;
	      }
	    }
	
	}
      
      /* Check depth buffer */
      glXGetConfig(glw_display, this_vis, GLX_DEPTH_SIZE, &val);
      if ( GLW_HAS_DEPTH(type) ) 
	{
	  if ( val <= 0 )
	    {
	      if (debug) fprintf(stderr,"no depth buffer support.\n");
	      continue;
	    }
	} 
      else
	{
	  if ( val > 0 )
	    {
	      if (debug) fprintf(stderr,"depth buffer support not demanded.\n");
	      continue;
	    }
	}
      
      /* Check stencil buffer */
      glXGetConfig(glw_display, this_vis, GLX_STENCIL_SIZE, &val);
      if ( GLW_HAS_STENCIL(type) )
	{
	  if ( val <= 0 )
	    {
	      if (debug) fprintf(stderr,"no stencil buffer support.\n");
	      continue;
	    }
	} 
      else
	{
	  if ( val > 0 )
	    {
	      if (debug) fprintf(stderr,"stencil buffer support not demanded.\n");
	      continue;
	    }
	}
      
      glXGetConfig(glw_display, this_vis, GLX_BUFFER_SIZE, &this_score);
      
      if (debug) fprintf(stderr,"ok, buffer size=%d.\n",this_score);
      if (this_score > best_score ) 
	{
	  best_score = this_score;
	  best_vis = this_vis;
	}
      
    }
  
  if ( best_vis ) 
    {
      sampleVis.visualid = best_vis->visualid;
      sampleVis.screen = glw_screen;
      if ( nvis > 0 ) XFree(vis_list);
      return XGetVisualInfo(glw_display, VisualIDMask|VisualScreenMask, 
			    &sampleVis, &nvis);
    } 
  else 
    {
      if ( nvis > 0 ) XFree(vis_list);
      return None;
    }
}

/*
old code 

static XVisualInfo *FindMainVisual(glWindow w,GLenum type)
{
  int list[32], i;
  i = 0;
  
  list[i++] = GLX_LEVEL;
  list[i++] = 0;
  
  if (GLW_IS_DOUBLE(type)) 
    {
      list[i++] = GLX_DOUBLEBUFFER;
    }

  if (GLW_IS_RGB(type)) 
    {
      list[i++] = GLX_RGBA;
      list[i++] = GLX_RED_SIZE;
      list[i++] = 1;
      list[i++] = GLX_GREEN_SIZE;
      list[i++] = 1;
      list[i++] = GLX_BLUE_SIZE;
      list[i++] = 1;

      if (GLW_HAS_ALPHA(type)) 
	{
	  list[i++] = GLX_ALPHA_SIZE;
	  list[i++] = 1;
	}
      
      if (GLW_HAS_ACCUM(type)) 
	{
	  list[i++] = GLX_ACCUM_RED_SIZE;
	  list[i++] = 1;
	  list[i++] = GLX_ACCUM_GREEN_SIZE;
	  list[i++] = 1;
	  list[i++] = GLX_ACCUM_BLUE_SIZE;
	  list[i++] = 1;
	  if (GLW_HAS_ALPHA(type)) {
	    list[i++] = GLX_ACCUM_ALPHA_SIZE;
	    list[i++] = 1;
	  }
	}
    }
  else if (GLW_IS_INDEX(type)) 
    {
      list[i++] = GLX_BUFFER_SIZE;
      list[i++] = 1;
    }
  
  if (GLW_HAS_DEPTH(type)) 
    {
      list[i++] = GLX_DEPTH_SIZE;
      list[i++] = 1;
    }
  
  if (GLW_HAS_STENCIL(type)) 
    {
      list[i++] = GLX_STENCIL_SIZE;
      list[i++] = 1;
    }
  
  list[i] = (int)None;
  return glXChooseVisual(glw_display, glw_screen, list);
}
*/

static XVisualInfo *FindOverlayVisual(glWindow w)
{
    int list[3];

    list[0] = GLX_LEVEL;
    list[1] = 1;
    list[2] = (int)None;
    return glXChooseVisual(glw_display, glw_screen, list);
}

static GLenum GetMainWindowType(glWindow w,XVisualInfo *vi)
{
    GLenum mask;
    int x, y, z;

    mask = (GLenum)0;

    glXGetConfig(glw_display, vi, GLX_DOUBLEBUFFER, &x);
    if (x) {
	mask |= GLW_DOUBLE;
    } else {
	mask |= GLW_SINGLE;
    }

    glXGetConfig(glw_display, vi, GLX_RGBA, &x);
    if (x) {
	mask |= GLW_RGB;
	glXGetConfig(glw_display, vi, GLX_ALPHA_SIZE, &x);
	if (x > 0) {
	    mask |= GLW_ALPHA;
	}
	glXGetConfig(glw_display, vi, GLX_ACCUM_RED_SIZE, &x);
	glXGetConfig(glw_display, vi, GLX_ACCUM_GREEN_SIZE, &y);
	glXGetConfig(glw_display, vi, GLX_ACCUM_BLUE_SIZE, &z);
	if (x > 0 && y > 0 && z > 0) {
	    mask |= GLW_ACCUM;
	}
    } else {
	mask |= GLW_INDEX;
    }

    glXGetConfig(glw_display, vi, GLX_DEPTH_SIZE, &x);
    if (x > 0) {
	mask |= GLW_DEPTH;
    }

    glXGetConfig(glw_display, vi, GLX_STENCIL_SIZE, &x);
    if (x > 0) {
	mask |= GLW_STENCIL;
    }

    if (glXIsDirect(glw_display, w->cMain)) {
	mask |= GLW_DIRECT;
    } else {
	mask |= GLW_INDIRECT;
    }

    return mask;
}

static XVisualInfo *get_visual_from_id( Display *dpy, int scr,int id)
{
  XVisualInfo temp,*xtemp;
  long mask;
  int n;
  int default_depth;
  int default_class;
  
   mask = VisualIDMask;
 
    temp.visualid = id;
  
  xtemp= XGetVisualInfo( dpy, mask, &temp, &n );
  return xtemp;

}


static int WaitForMainWindow(Display *d, XEvent *e, char *arg)
{
  glWindow wEvent=(glWindow)arg;
  if (e->type == MapNotify && e->xmap.window == wEvent->w_main) {
    return GL_TRUE;
  } else {
    return GL_FALSE;
  }
}

/*
static int WaitForOverlayWindow(Display *d, XEvent *e, char *arg)
{

    if (e->type == MapNotify && e->xmap.window == wEvent->w_overlay) {
	return GL_TRUE;
    } else {
	return GL_FALSE;
    }
    return GL_FALSE;
}
*/
static void default_expose_func(glWindow w, 
			 void *user_data,
			 int width, 
			 int height)
{
  
  /* force redraw if no user drivers set */

}

static void create_window(glWindow w, int overlayFlag)
{
  XSetWindowAttributes wa;
  XTextProperty tp;
  XSizeHints sh;
  XEvent e;
 
  wa.colormap = w->cMapMain;
  wa.background_pixmap = None;
  wa.border_pixel = 0;
  wa.backing_store=Always;
  wa.save_under=True;
  wa.event_mask = StructureNotifyMask | ExposureMask | KeyPressMask |
    ButtonPressMask | ButtonReleaseMask | PointerMotionMask;

  w->w_main = XCreateWindow(glw_display, glw_root, w->x, w->y, w->w, w->h, 0,
			   w->vInfoMain->depth, InputOutput,
			   w->vInfoMain->visual,
			   CWBackPixmap|CWBorderPixel|CWEventMask|CWColormap,
			   &wa);
  
  XStringListToTextProperty(&w->title, 1, &tp);
  sh.flags = USPosition | USSize | PMaxSize |PResizeInc;
  sh.width_inc=16;
  sh.height_inc=16;
  if (w->keep_aspect)
    {
      
      sh.flags|=PAspect;
      sh.min_aspect.x=w->w;
      sh.min_aspect.y=w->h;
      
      sh.max_aspect.x=w->w;
      sh.max_aspect.y=w->h;
    }
  
  {
    Screen *scr = DefaultScreenOfDisplay(glw_display);
    sh.max_width=WidthOfScreen(scr);
    sh.max_height=HeightOfScreen(scr);
  }
  
  if (w->w_toplevel!=0) 
    XSetTransientForHint(glw_display, w->w_main, w->w_toplevel);
  XSetWMProperties(glw_display, w->w_main, &tp, &tp, 0, 0, &sh, 0, 0);
  w->deleteWindowAtom = XInternAtom(glw_display, "WM_DELETE_WINDOW", False);
  XSetWMProtocols(glw_display, w->w_main, &w->deleteWindowAtom, 1);
  glwTrace(w,"mapwindow");
  XMapWindow(glw_display, w->w_main);
  /*  w->drawAllowFlag = GL_FALSE;*/
  w->lastEventType= -1;
  glwTrace(w,"waitfomain");
  XSync(glw_display,False);
  XPeekIfEvent(glw_display, &e, WaitForMainWindow, (XPointer)w);
  XSync(glw_display,False);
  glwTrace(w,"waitfomain");

  if (overlayFlag == GL_TRUE) {
    w->vInfoOverlay = FindOverlayVisual(w);
    if (w->vInfoOverlay) {
      w->cOverlay = glXCreateContext(glw_display, w->vInfoOverlay, None,
				     GL_TRUE);
      w->cMapOverlay = XCreateColormap(glw_display, glw_root,
				       w->vInfoOverlay->visual, AllocNone);
      glwSetOverlayMap(w,256, colorMaps);
      wa.colormap = w->cMapOverlay;
      wa.background_pixmap = None;
      wa.border_pixel = 0;
      w->w_overlay = XCreateWindow(glw_display, w->w_main, 0, 0, w->w, w->h, 0,
				  w->vInfoOverlay->depth, InputOutput,
				       w->vInfoOverlay->visual,
				  CWBackPixmap|CWBorderPixel|CWColormap,
				  &wa);
      XMapWindow(glw_display, w->w_overlay);
      XSetWMColormapWindows(glw_display, w->w_main, &w->w_overlay, 1);
      w->type |= GLW_OVERLAY;
    } else {
      glwError("Unable to create overlay plane!");
    }
  }
  glwTrace(w,"cursor")
    /*  w->wait_cursor=XCreateFontCursor(glw_display,XC_watch);
	w->user_cursor=XCreateFontCursor(glw_display,XC_target);*/
  /*  XDefineCursor(glw_display,w->w_main, w->user_cursor);*/
}

static void create_pixmap(glWindow w, int overlayFlag)
{
  w->pixmap = XCreatePixmap(glw_display, 
			    glw_root,
			    w->w, 
			    w->h, 
			    w->vInfoMain->depth);
  if (w->pixmap==0)
    glwError("Unable to create pixmap!");
  else
    glwTrace(w,"pixmap ok");

  w->glx_pixmap =glXCreateGLXPixmap(glw_display,
				   w->vInfoMain,
				   w->pixmap);
  if (w->glx_pixmap==0)
    glwError("Unable to create glx_pixmap!");
  else
    glwTrace(w,"glxpixmap ok");
    
}

void glwDummy(glWindow w, void * data){;}


static glWindow glw_xcreate(glWindow w)
{
  GLenum overlayFlag;
  glwTrace(w,"xcreate");
 
  
 
  if (w->offscreen) w->type |= GLW_INDIRECT;
  
  if (GLW_HAS_OVERLAY(w->type))
      w->overlayFlag = GL_TRUE;
  else 
      w->overlayFlag = GL_FALSE;
  w->type &= ~GLW_OVERLAY;
  
  
  if (w->dm_policy == GLW_MINIMUM_CRITERIA)
    w->vInfoMain = FindBestMainVisual(w,w->type);
  else if (w->dm_policy == GLW_EXACT_MATCH)
    w->vInfoMain = FindExactMainVisual(w,w->type);

  glwTrace(w,"xcreate");

  if (!w->vInfoMain)
    {
      char*vid;
      if (vid=getenv("GLTOOLS_VISUAL"))
	 w->vInfoMain=get_visual_from_id(glw_display, glw_screen,atoi(vid));
      if (!w->vInfoMain)
	{
	  glwError("Main visual not found!");
	  exit(1);
	}
    }
  glwTrace(w,"xcreate");
  
  if (debug) fprintf(stderr,"glwin: using  visual id 0x%x\n", (unsigned)w->vInfoMain->visualid);
  
#if defined(__cplusplus) || defined(c_plusplus)
  w->vInfoMainClass = w->vInfoMain->c_class;
#else
  w->vInfoMainClass = w->vInfoMain->class;
#endif  
  
  
  w->cMain = glXCreateContext(glw_display, w->vInfoMain, None,
			      (GLW_IS_DIRECT(w->type))?GL_TRUE:GL_FALSE);
  if (!w->cMain) 
    {
      glwError("Unable to create glx context!");
      exit(1);
    }
  
  w->type = GetMainWindowType(w,w->vInfoMain);
  
  if (GLW_IS_INDEX(w->type)) {
    if (w->vInfoMainClass != StaticColor &&
	w->vInfoMainClass != StaticGray) {
      w->cMapMain = XCreateColormap(glw_display, glw_root, w->vInfoMain->visual,
				    AllocAll);
    } else {
      w->cMapMain = XCreateColormap(glw_display, glw_root, w->vInfoMain->visual,
				    AllocNone);
    }
  } else {
    
    /* alt ging schief fuer MESA auf Xterm 
       w->cMapMain = XCreateColormap(glw_display, glw_root, w->vInfoMain->visual,
       AllocNone);
    */	
    /* RGB colormap is AllocNone, share the root colormap if possible */
    Screen *scr = DefaultScreenOfDisplay(glw_display);
    if (MaxCmapsOfScreen(scr)==1
	&& w->vInfoMain->visual==DefaultVisual(glw_display,glw_screen)) {
      w->cMapMain = DefaultColormap(glw_display,glw_screen);
    }
    else {
      w->cMapMain = XCreateColormap(glw_display, glw_root, w->vInfoMain->visual,
				    AllocNone);
    }
  }
  /* PRE-MESA
     glwSetRGBMap(w,256, colorMaps);
  */   
  
  if (GLW_IS_INDEX(w->type) || w->vInfoMainClass==DirectColor) {
    glwSetRGBMap(w,256, colorMaps);
  }
  
  if (w->offscreen)
    create_pixmap(w,w->overlayFlag);
  else
    create_window(w,w->overlayFlag);
  

  glwSetExposeFunc(w,default_expose_func);
  glwTrace(w,"fonts");
  glwSetFontSize(w,GLW_DEFAULT_FONT_SIZE);
  glwLoadFont(w,w->font_to_use);
  glwTrace(w,"window ok");

  w->create_request=0;
  glwTrace(w,"xcreate ok");
  return w;
}

glWindow glwCreate(void)
{
  glWindow w;
  w=glw_init_window();
  glw_init_reset();
  glw_insert(w);
  w->redraw_func=NULL;
  w->application_redraw_request=1;
  if (!glw_threads)
    glw_main_loop();
  glwTrace(w,"glwCreate");
  return w;
}

void glwLock(void)
{
#ifdef PTHREADS
  if (glw_threads)
    assert(pthread_mutex_lock(&glw_mutex)==0);
#endif
  
}

void glwUnlock(void)
{
#ifdef PTHREADS
  if (glw_threads)
    pthread_mutex_unlock(&glw_mutex);
#endif
  
}

static void update_title(glWindow w)
{
  XTextProperty tp;
  if (w->offscreen) return;
  if (w->state!=NULL) 
    {
      sprintf(w->xtitle,"%s [ %s ]",w->title,w->state);
      XStringListToTextProperty(&w->xtitle, 1, &tp);
    }
    else
      XStringListToTextProperty(&w->title, 1, &tp);
  
  XSetWMProperties(glw_display, w->w_main, &tp, 0, 0, 0, 0, 0, 0);
  XFlush(glw_display);
}

void glwShowState(glWindow w, char *state)
{
 if (state!=NULL)
   { 
     if (w->state==NULL)
       w->state=(char*)malloc(GLW_NAMELEN);
     strcpy(w->state,state);
   }
 else
  {
    free(w->state);
    w->state=NULL;
  }
 update_title(w);
}

void glwSetTitle(glWindow w, char *title)
{
  strcpy(w->title,title);
  update_title(w);
}

char * glwGetTitle(glWindow w)
{
  return w->title;
}

#endif
/*
 * $Log: glwin.c,v $
 * Revision 2.84  2003/03/28 16:47:48  fuhrmann
 * pdelib1.15_alpha1
 *
 * Revision 2.83  2003/03/28 11:20:26  fuhrmann
 * pdelib2.0_alpha1
 *
 * Revision 2.82  2003/02/25 16:09:56  fuhrmann
 * don't set debug by default...
 *
 * Revision 2.81  2003/02/20 16:50:45  fuhrmann
 * removed license constraints
 *
 * Revision 2.80  2003/01/15 09:19:31  fuhrmann
 * *** empty log message ***
 *
 * Revision 2.79  2003/01/14 13:18:59  fuhrmann
 * event bug removed
 *
 * Revision 2.78  2003/01/10 11:04:23  fuhrmann
 * checked some of hang's problems
 *
 * Revision 2.77  2002/09/24 08:55:44  fuhrmann
 * scales
 *
 * Revision 2.76  2002/09/23 09:18:45  fuhrmann
 * scale hacks for gleps,glrnd
 *
 * Revision 2.75  2002/07/01 12:00:13  fuhrmann
 * window startup stuff improved
 *
 * Revision 2.74  2002/06/22 13:42:01  fuhrmann
 * interim
 *
 * Revision 2.73  2002/06/20 15:13:25  fuhrmann
 * threaded  gltools
 *
 * Revision 2.72  2002/06/20 14:47:11  fuhrmann
 * pthreads seem to run
 *
 * Revision 2.71  2002/06/20 09:40:54  fuhrmann
 * one event loop for all windows
 *
 * Revision 2.70  2002/06/19 17:44:30  fuhrmann
 * interim
 *
 * Revision 2.69  2002/06/18 14:27:39  fuhrmann
 * interim
 *
 * Revision 2.68  2002/06/17 17:20:20  fuhrmann
 * glw_threaded gltools ok on linux & tru64
 *
 * Revision 2.67  2002/06/17 15:33:11  fuhrmann
 * first working revision with pglw_threads & XSendEvent
 *
 * Revision 2.66  2002/06/14 17:25:06  fuhrmann
 * glw_threading: interim version
 *
 * Revision 2.65  2002/06/13 15:32:21  fuhrmann
 * pthreads in glrnd
 *
 * Revision 2.64  2002/01/21 10:15:56  fuhrmann
 * font change
 *
 * Revision 2.63  2002/01/18 09:12:00  fuhrmann
 * running test fit
 *
 * Revision 2.62  2002/01/08 15:48:38  fuhrmann
 * revert to fltk 1.0.11
 *
 * Revision 2.61  2001/07/27 17:12:22  fuhrmann
 * mpfit stuff
 *
 * Revision 2.60  2001/07/25 17:10:03  fuhrmann
 * gldigi
 *
 * Revision 2.59  2001/07/23 12:14:25  fuhrmann
 * Juelich Juli 2001
 *
 * Revision 2.58  2001/06/01 16:08:12  fuhrmann
 * *** empty log message ***
 *
 * Revision 2.57  2001/05/04 15:19:17  fuhrmann
 * pardiso preconditioner
 *
 * Revision 2.56  2000/12/10 23:04:44  fuhrmann
 * macro order corrected
 *
 * Revision 2.55  2000/12/10 22:58:09  fuhrmann
 * gleps bug
 * flglwin proto works
 *
 * Revision 2.54  2000/12/08 16:39:43  fuhrmann
 * first steps to fltk
 *
 * Revision 2.53  2000/11/22 15:20:49  fuhrmann
 * ICE hack
 * 1 mpeg/eps per variation
 * variation[i]-> variationi
 * bias=... option
 *
 * Revision 2.52  2000/11/20 11:10:23  fuhrmann
 * eps graphics stuff
 *
 * Revision 2.51  2000/10/25 10:49:59  fuhrmann
 * crop stuff for graphics
 *
 * Revision 2.50  2000/10/19 09:54:59  fuhrmann
 * scl graphics stuff
 *
 * Revision 2.49  2000/07/27 12:43:36  schmelze
 * *** empty log message ***
 *
 * Revision 2.48  2000/03/06 14:54:13  fuhrmann
 * removed toplevel info output
 *
 * Revision 2.47  2000/02/22 17:27:06  fuhrmann
 * * transient_for in glwin
 * * 0.1a Version of WIASTwoPhase
 *
 * Revision 2.46  2000/01/13 17:36:41  fuhrmann
 * optmized ps output
 *
 * Revision 2.45  2000/01/11 16:33:41  fuhrmann
 * name buffer sizes
 *
 * Revision 2.44  1999/11/12 19:50:11  fuhrmann
 * no abort for X errors
 *
 * Revision 2.43  1999/11/10 12:17:21  fuhrmann
 * sysconlaw.lua, working solve/replay stuff, Append protocol etc.
 *
 * Revision 2.42  1999/11/04 17:16:09  philip
 * string memory bug in glwin
 * roundoff error in colors in gleps
 *
 * Revision 2.41  1999/11/04 16:43:40  philip
 * removed calloc bug
 *
 * Revision 2.40  1999/11/01 18:31:24  fuhrmann
 * some problem with glXmakecurrent
 *
 * Revision 2.39  1999/09/17 16:35:01  fuhrmann
 * feedback stuff war kaputtrepariert...
 *
 * Revision 2.38  1999/09/15 16:16:24  fuhrmann
 * size_t for feedback buffer
 * file !=NULL check for dump files
 *
 * Revision 2.37  1999/08/19 15:31:46  fuhrmann
 * strange memory leak in glwDestroy !!! STILL THERE !!!
 *
 * Revision 2.36  1999/07/02 10:38:05  fuhrmann
 * Mesa bug neutralized, gleps reference taken out of glwin
 *
 * Revision 2.35  1999/06/30 13:07:50  fuhrmann
 * font size stuff
 *
 * Revision 2.34  1999/06/30 10:55:45  fuhrmann
 * no exit at token errors
 *
 * Revision 2.33  1999/06/30 10:27:25  fuhrmann
 * some bells &whistles in feedback stuff
 *
 * Revision 2.32  1999/06/29 17:56:24  fuhrmann
 * Feedback buffer is used now...
 *
 * Revision 2.31  1999/04/20 17:32:16  fuhrmann
 * sgi warnings, 3D flux test
 *
 * Revision 2.30  1999/04/13 11:43:31  fuhrmann
 * more thorough checks, k&r stuff deleted
 *
 * Revision 2.29  1999/04/08 15:03:15  fuhrmann
 * glmesh split up
 *
 * Revision 2.28  1998/09/04  15:25:48  fuhrmann
 * skip frame option in StartRecording
 *
 * Revision 2.27  1998/07/09  14:39:52  fuhrmann
 * initialization bug in mpeg stuff removed
 *
 * Revision 2.26  1998/07/06  10:45:12  fuhrmann
 * Increased compiler warning level for all PDELIB-1.7 modules,
 * removed all bugs, some warnings remain (rcsids, ld_init, unused parameters)
 *
 * Revision 2.25  1998/05/18  17:19:27  fuhrmann
 * if file given in mpeg param ressource is not existent then invoke
 * default mechanism.
 *
 * Revision 2.24  1998/03/11  19:21:04  fuhrmann
 * Hey we can create batch videos! (still beta, but...)
 *
 * Revision 2.23  1998/03/06  16:45:31  fuhrmann
 * tiff->ppm
 * print command
 * video recording
 *
 * Revision 2.22  1998/03/05  20:01:24  fuhrmann
 * interim test version with SGI movie lib : it was a mess...
 *
 * Revision 2.21  1998/01/26  11:07:28  schmelze
 * tests from external working directory, box*test.html
 *
 * Revision 2.20  1998/01/14  18:27:34  fuhrmann
 * Merge mit Ilja
 *
 * Revision 2.19  1997/11/27  19:03:09  fuhrmann
 * glwRecord stuff, PAL-Format, tex-file for keys
 *
 * Revision 2.18  1997/11/04  15:53:30  fuhrmann
 * color scales ; crash with glrnd
 *
 * Revision 2.17  1997/10/17  16:54:58  fuhrmann
 * glwFlush
 *
 * Revision 2.16  1997/10/01  15:41:27  fuhrmann
 * GLW_MOD1
 *
 * Revision 2.15  1997/09/09  17:42:28  fuhrmann
 * cursor stuff
 *
 * Revision 2.14  1997/04/15  08:10:22  fuhrmann
 * still not clear which font to load...
 *
 * Revision 2.13  1997/03/20  18:19:18  fuhrmann
 * glwGrab in no-tiff case
 *
 * Revision 2.12  1997/02/28  17:41:13  fuhrmann
 * glwGrab code
 *
 * Revision 2.11  1997/01/27  14:30:45  fuhrmann
 * other default font
 *
 * Revision 2.10  1996/10/09  16:33:57  fuhrmann
 * keys next, prev, tab
 *
 * Revision 2.9  1996/10/02  11:12:05  fuhrmann
 * some debugging in visual stuff, no change of algorithm
 *
 * Revision 2.8  1996/09/23  17:47:27  fuhrmann
 * -DUSE_OS_MESA for mesa offscreen contexts
 *
 * Revision 2.7  1996/09/19  16:57:34  fuhrmann
 * typo in glwProcess removed
 *
 * Revision 2.6  1996/09/19  16:33:39  fuhrmann
 * GLW_DEPTH in default mode
 * redraw first when enterinh glw_process in event mode
 *
 * Revision 2.5  1996/09/03  09:22:41  fuhrmann
 * tiffio.h location changed
 *
 * Revision 2.4  1996/09/03  09:09:04  fuhrmann
 * - fonts & printing characters
 * - keep aspect ratio
 * - dump pixmap size can default to window size
 *
 * Revision 2.3  1996/08/30  10:47:34  fuhrmann
 * Default expose function set (glw should redraw in event driven
 * mode if user has no callbacks set)
 *
 * Revision 2.2  1996/04/17  16:47:56  fuhrmann
 * error in off-screen rendering: PixelAlignment in GL buf was wrong
 * error in off-screen rendering: wrong count of lines (black upper line in GL dump) fixed
 *
 * Revision 2.1  1996/03/12  09:49:27  fuhrmann
 * Off-screen rendering into memory with large pixmaps using osmesa
 *
 * Revision 2.0  1996/02/15  20:26:14  fuhrmann
 * First meta-stable distribution
 *
 * Revision 1.9  1996/02/15  20:25:00  fuhrmann
 * - Error in return value for glwDump removed
 *
 * Revision 1.8  1996/02/15  14:18:20  fuhrmann
 * - improved behaviour when tiff file not opened
 *
 * Revision 1.7  1996/02/09  16:45:02  fuhrmann
 * - shift keys now part of button mask
 * - more keysyms available
 *
 * Revision 1.6  1996/02/06  09:59:46  fuhrmann
 * - glwDump ( depends on libtiff.a !)
 * - glwSetIdleFunc and glwSetDisplayFunc obsolete,
 *   use glwSetRedrawFunc & glwSetControlMode (better to understand)
 *
 * Revision 1.5  1996/01/30  18:36:15  fuhrmann
 * Changed default title to gltools-1.0
 *
 * Revision 1.4  1995/11/08  12:05:39  fuhrmann
 * glwSetTitle introduced
 *
 * Revision 1.3  1995/10/24  18:34:15  fuhrmann
 * Namelen von 64 auf 128  ???
 * Merge mit libtk 1.1 (dm_policy stuff) vielleicht unnoetig...
 *
 * Revision 1.2  1995/10/24  12:40:51  fuhrmann
 * Zeilennr in glwError
 *
 * Revision 1.1  1995/10/20  15:42:03  fuhrmann
 * Initial revision
 *
 */
